def output_help(out)
	out.puts "\n======================================================================================================"
	out.puts "Randomness is due to a) the tuning algorithm itself, b) Sampled instances, and c) different seeds.\n"
	
	out.puts "\nParameters"
	out.puts "=========="
	
	out.puts "-instanceFile <X>, where <X> is the path to a text file with instances (see README.TXT for details)"
	out.puts "-ordered <X>, where <X> in {0,1} [#{@ordered_instances}]. When this is 1, preserve the order of re-sampled instances (e.g. easy first)."
	out.puts "-numRepetitions <X>, where <X> is a positive integer [#{@numRepetitions}]"
	out.puts "-deterministic <X>, where <X> is in {0,1}[#{@deterministic_algorithm}]"
	out.puts "-stratified <X>, where <X> is in {0,1}[#{@stratified}]"
end

# =========================================
#                  MAIN
# =========================================

@instanceset_filename = "sw-instances.txt"
@numRepetitions = 100
@deterministic_algorithm = 0
@stratified = 1

if ARGV.length < 1
	output_help($stdout)
	exit
end

# =========================================
# Read in command line options.
# =========================================
0.step(ARGV.length-1, 2){|i|
	case ARGV[i]
	when "-instanceFile"
		@instanceset_filename = ARGV[i+1]
	when "-stratified"
		@stratified = ARGV[i+1]
	when "-numRepetitions"
		@numRepetitions = ARGV[i+1].to_i
	when "-deterministic"
		@deterministic_algorithm = ARGV[i+1]
	else
		output_help($stdout)
		puts "\n\n"
		raise "Unknown argument: #{ARGV[i]}"
	end
}
 
@deterministic_algorithm= false if @deterministic_algorithm == "0" or @deterministic_algorithm == 0
@stratified = false if @stratified == "0" or @stratified == 0

orig_instances_sorted = []
File.open(@instanceset_filename){|file|
	while line = file.gets
		orig_instances_sorted << line.chomp
	end
}

# =========================================
# Set up the seeded instance files.
# =========================================
@det_str = ""
@det_str = "-det_algo" if @deterministic_algorithm

@strat_str = ""
@strat_str = "-strat" if @stratified

subdir = "#{@instanceset_filename.sub(/\.txt/,"")}#{@strat_str}#{@det_str}#{@order_str}"
system "mkdir #{subdir}" unless File.exist?(subdir)

seed = 1234
srand(seed)

#=== @numRepetitions times, take 2000 samples <instance, seed>, with replacement.
for repetition in 0...@numRepetitions 
	@inst_seeds = []
	strat_instances = []
	for i in 0...2000
		if @stratified
			strat_instances = orig_instances_sorted.dup if strat_instances.empty?
			inst = strat_instances[rand(strat_instances.length)]
			strat_instances.delete(inst)
		else
			inst = orig_instances_sorted[rand(orig_instances_sorted.length)]
		end
		algoseed = rand(2147483646)+1 	# Let's not use zero, some algos don't accept that as a seed.
		algoseed = -1 if @deterministic_algorithm
		@inst_seeds << [inst, algoseed]
	end

	filename = "#{subdir}/instance_seed_file_#{repetition}.txt"
	File.open(filename, "w"){|file|
		for inst_seed in @inst_seeds
			inst, algoseed = inst_seed
			file.puts "#{algoseed} #{inst} 0"
		end
	}
end