#Methods that are used in multiple scripts.
#require "param_ils_helper.rb" # just for tmpdir

require "algo_specifics.rb"

def tmpdir()
	return "/tmp"
end

def float_regexp()
	return '[+-]?\d+(?:\.\d+)?(?:[eE][+-]\d+)?'; 
end

unless $work_dir
	File.popen("pwd"){|file| $work_dir = file.gets.chomp}
end
$script_dir_for_cluster ="/.autofs/csother/ubccsprojectarrow/hutter/ParamILS/scripts/"


###########################################################################
#Perform a 2-sided Wilcoxon paired sign rank test. Return value: [p_value, median < 0]
###########################################################################
def wilcoxon_test(alpha, x, y)
	minlen = [x.length, y.length].min
	
	#=== With less than six observations you can't determine significance, so return right away.
	return [false, -1] if minlen < 6
	
	#=== Compute difference of the two vectors.
	diff = []
	for i in 0...minlen
		diff[i] = x[i] - y[i]
	end

	#=== Construct the file holding the R script.
	r_command_filename = "/tmp/tmp_r_cmd_#{random_number_without_rand}"
	r_result_filename = "/tmp/tmp_r_result_#{random_number_without_rand}"
	array_string = diff.join(", ")
	File.open(r_command_filename, "w"){|f|
		f.puts "diff<-c(#{array_string})"
		f.puts "p_value<-wilcox.test(diff, exact=FALSE)$p.value"
		f.puts "write(p_value, '#{r_result_filename}')"
	}

	#=== Call the R script and read the results from the result file.
	system "R CMD BATCH #{r_command_filename}"
	File.delete(r_command_filename)
	pval = -1
	File.open(r_result_filename){|f| pval = f.gets.chomp.to_f}
	File.delete(r_result_filename)
	dir, cmd_filename = File.split(r_command_filename)
	File.delete(cmd_filename + ".Rout")
	
	print diff.sort{|x,y| x.abs <=> y.abs}.map{|x| x<0 ? "-" : (x==0 ? "0" : "+" )}.join(" ")
	puts "  #{pval}"

	require "stats_ils.rb"
#	puts "Testing [#{x.join(", ")}] vs [#{y.join(", ")}]: pval=#{pval}"
	return [pval < alpha/2, pval]  #two-sided test, so divide alpha by 2.
end

###########################################################################
#Sample from an exponential distribution.
###########################################################################
def generate_samples_from_exponential(median, numSamples)
#=== Simple ruby wrapper that takes the mean of an exponential distribution and generates a sample by calling a built-in function in python.
#One sample works from the command line:	cmd = "python -c \"import random; print random.expovariate(#{1/mean})\""
	mean = 1/Math.log(2) * median
	cmd = "python ../scripts/exponential_samples.pyt #{1/mean} #{numSamples}"
	results = []
	File.popen(cmd){|f| results = f.readlines}
	return results.map{|x|x.chomp.to_f}
end


###########################################################################
# Run a set of commands on the cluster.
###########################################################################
def runCommandsOnCluster(commands, waitToFinish=false, name = "")
	t=Time.now
	datetime = t.strftime("%Y-%m-%d %H-%M-%S") # YYYY-MM-DD HH:MM:SS
	todoFilename = "#{tmpdir}/#{name}-tmpAlgosToRun-#{datetime}-#{random_number_without_rand}".gsub(/ /,"")
	File.open(todoFilename, "w"){|f| 
		for command in commands
			f.puts "#{command}"
		end
	}

	shFilename = "#{tmpdir}/#{name}-tmpShFile-#{datetime}-#{random_number_without_rand}.sh".gsub(/ /,"")
	#puts "todoFilename: #{todoFilename}, shFilename: #{shFilename}"
	File.open(shFilename, "w"){|f|
		f.puts "#!/bin/sh"
		f.puts "echo \"Here's what we know from the SGE environment\""
		f.puts "echo HOME=$HOME"
		f.puts "echo USER=$USER"
		f.puts "echo JOB_ID=$JOB_ID"
		f.puts "echo JOB_NAME=$JOB_NAME"
		f.puts "echo HOSTNAME=$HOSTNAME"
		f.puts "echo SGE_TASK_ID=$SGE_TASK_ID"
		f.puts "echo STDOUT_FILE=$stdout"
		f.puts "echo STDERR_FILE=$stderr"
		f.puts "echo jobin=$jobin"
		f.puts "echo jobout=$jobout"
		f.puts "echo joberr=$joberr"
		f.puts "RUBYLIB=/ubc/cs/home/h/hutter/arrowspace/ParamILS/scripts:/cs/public/lib/pkg/ruby-1.8.2/lib/ruby/site_ruby/1.8:/.autofs/binaries/cspubliclib/pkg/ruby-mysql/mysql-ruby-2.7.1"
		f.puts "export RUBYLIB"
		#/cs/local/bin in the PATH is absolutely necessary for JAVA, get java.lang.NoClassDefFoundError otherwise !!!
		f.puts "PATH=/cs/beta/lib/pkg/sge-6.0u7_1/bin/lx24-x86:/cs/local/bin:/cs/local/generic/bin:/cs/local/bin/pbm+:/usr/local/bin:/usr/bin/X11:/bin:/usr/ucb:/usr/bin:/ubc/cs/home/h/hutter/bin:/ubc/cs/home/h/hutter/bin/ix86linux:/opt/kde3/bin:/opt/gnome/bin:/usr/games:/usr/sbin:/sbin:/usr/lib/java/bin:/cs/public/bin:/cs/public/generic/bin:/cs/public/bin/xwindows:/ubc/cs/home/h/hutter/ruby-scripts:/ubc/cs/home/h/hutter/ruby-scripts/autoparam/:/ubc/cs/home/h/hutter/bioinf/impl/scripts:/ubc/cs/home/h/hutter/lib:/ubc/cs/home/h/hutter/mcmc/bin/:/ubc/cs/home/h/hutter/arrowspace/ParamILS/testbed/scripts:/ubc/cs/home/h/hutter/arrowspace/ParamILS/scripts:/cs/public/lib/pkg/ruby-1.8.2/lib/ruby/site_ruby/1.8:/.autofs/binaries/cspubliclib/pkg/ruby-mysql/mysql-ruby-2.7.1:/ubc/cs/home/h/hutter/ant/bin:$SGE_BINARY_PATH:$PATH"
		f.puts "export PATH"
		f.puts "#\$ -o /ubc/cs/home/h/hutter/arrowspace/sgeout -e /ubc/cs/home/h/hutter/arrowspace/sgeout"
#		f.puts "#\$ -o /dev/null -e /dev/null"
		f.puts "line=`head -n $SGE_TASK_ID #{todoFilename} | tail -1`	# Get line of todoFilename."

		cmd = "cd #{$work_dir};"
		f.puts "echo Calling: #{cmd}  #output"
		f.puts cmd

		#cmd= "pwd"
		#f.puts "echo Calling: #{cmd}  #output"
		#f.puts cmd
		cmd= "$line"
		f.puts "echo Calling: #{cmd}  #output"
		f.puts cmd
	}
	
	if commands.length > 0
		consumables = ""
		consumables += " -l dbheavy=1" if $dbheavy
		consumables += " -l db=1" if $dblight
		consumables += " -l cplex=1" if $cplex
		sge_cmd = "qsub -cwd -t 1-#{commands.length} -l memheavy=1 -P eh #{consumables} #{shFilename}"
	end
	
	puts sge_cmd
	puts shFilename
	jobid = nil
	#=== Start job and remember job id.
	File.popen(sge_cmd){|sge_response|
		line = sge_response.gets
		puts line
		line =~ /Your job (\d+)\./
		jobid = $1.to_i
	}
##	File.delete(todoFilename)
#	File.delete(shFilename)

	if waitToFinish
	#=== Sleep until job done.
		puts "Waiting for SGE job #{jobid} to finish. TODO filename = #{todoFilename}"
		still_running = true
		while still_running
			sleep(100)
			still_running = false
			File.popen("qstat"){|qstat_output|
				while line = qstat_output.gets
					still_running = true if line =~ /^\s*#{jobid}\s*/
				end
			}
		end
	end
	return [jobid, todoFilename]
end


###########################################################################
# Run a set of algorithm runs the simple way (directly on the local machine).
###########################################################################
def runAlgosLocally(algosToRun)
	for algoToRun in algosToRun
		runalgo(algoToRun[0], algoToRun[1])
		#system "ruby #{$script_dir_for_cluster}/runalgo.rb #{algoToRun[0]} #{algoToRun[1]}"
	end
end

###########################################################################
# Run a set of algorithm runs on the cluster.
###########################################################################
def runAlgosOnCluster(algosToRun, waitToFinish=false, algo = "")
	commandsToRun = []
	for algoToRun in algosToRun
	#cd #{$work_dir}; 
		commandsToRun << "/usr/bin/ruby #{$script_dir_for_cluster}/runalgo.rb #{algoToRun[0]} #{algoToRun[1]}"
	end
	runCommandsOnCluster(commandsToRun, waitToFinish, algo)
end


###########################################################################
# Run a set of algorithm runs with the jobserver.
###########################################################################
def runAlgosWithJobserver(algosToRun)
	require 'jobserver'

	#=== Put jobs in the queque.
	myJobQueue = []
	for algoToRun in algosToRun
		cmd = "cd #{$work_dir}; /usr/bin/ruby #{$script_dir_for_cluster}/runalgo.rb #{algoToRun[0]} #{algoToRun[1]}"
		myJobQueue << Job.new(:name=>"ALGORUN_CONFIG #{algoToRun[0]}, seed #{algoToRun[1]}.", :client_command=>cmd)
	end

	if myJobQueue.length > 0
		#=== Configure jobserver.
		Job.nicelevel = nil # nice doesn't work at UBC
		server = JobServer.new(myJobQueue, $work_dir, 0)

		hosts =  %w{arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48 arrow49 arrow50}
#		hosts =  %w{arrow25}
#		hosts =  %w{arrow36 arrow37 arrow38 arrow39 arrow40 arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48 arrow49}
#		hosts =  %w{arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48 arrow49}
#		hosts =  %w{arrow21 arrow22 arrow24 arrow26 arrow27 arrow30 arrow31 arrow34 arrow36 arrow37 arrow38 arrow39 arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48 arrow49}
#		hosts =  %w{arrow43 arrow44 arrow45 arrow46 arrow47 arrow48 arrow49}
#		hosts =  %w{arrow01 arrow02 arrow03 arrow21 arrow22 arrow23 arrow24 arrow25 arrow27 arrow28 arrow29 arrow30 arrow31 arrow32 arrow33 arrow34 arrow35 arrow36 arrow37 arrow40 arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48}
#		hosts =  %w{arrow30 arrow31 arrow32 arrow33 arrow34 arrow35 arrow36 arrow37 arrow40 arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48}
#		hosts =  %w{arrow01 arrow02 arrow03 arrow04 arrow05 arrow06 arrow07 arrow08 arrow09 arrow10 arrow11 arrow12 arrow13 arrow14 arrow15 arrow16 arrow17 arrow18 arrow19 arrow20 arrow21 arrow22 arrow23 arrow24 arrow25 arrow26 arrow27 arrow28 arrow29 arrow30 arrow31 arrow32 arrow33 arrow34 arrow35 arrow36 arrow37 arrow38 arrow39 arrow40 arrow41 arrow42 arrow43 arrow44 arrow45 arrow46 arrow47 arrow48 arrow49}
		
		while not hosts.empty? # Add workers in randomized order, so not everyone pounds on the last worker.
			randind = (random_number_without_rand*hosts.length).floor
			#randind = rand(hosts.length)
			host = hosts[randind]
			hosts.delete_at(randind)
			server.add_ssh_worker(host, $work_dir, 2)
		end

		#=== Dump out statistics on the progress every 5 seconds. 
		server.dumpStatistics #($statsfile, 5)

		#=== Wait until all jobs have finished.
		server.serve
	end
end


###########################################################################
# Gets the results for running algo on multiple instances - keeps track of which runs are used and does not reuse them !
###########################################################################
def getAlgoResultsForInstsAndParams(algo, instanceAndParamsHash, cutoff_time, cutoff_length, sorted_instances, oncluster=0, db=false)
#	puts "getAlgoResultsForInstsAndParams begin"
#	$stdout.flush

	unless db
		#=== Do runs, save in instanceAndParamsHash, return
		for instance in instanceAndParamsHash.keys
			for param_string in instanceAndParamsHash[instance].keys
				params = instanceAndParamsHash[instance][param_string]["params"]
				qual = instanceAndParamsHash[instance][param_string]["qual"]
				reference = instanceAndParamsHash[instance][param_string]["reference"]
				seeds = instanceAndParamsHash[instance][param_string]["seeds"]

				instanceAndParamsHash[instance][param_string]["results"] = {}
				for seed in seeds
					t = Time.now
					datetime = t.strftime("%Y-%m-%d %H:%M:%S") # YYYY-MM-DD HH:MM:SS
					algo_output_file = "#{tmpdir}/tmp-#{datetime}-#{random_number_without_rand}".gsub(/ /,"")

					paramstring = params.keys.map{|x| "-#{x} #{params[x]}"}.join(" ")
					cmd = "#{algo} #{instance} #{qual} #{cutoff_time} #{cutoff_length} #{seed} #{paramstring} > #{algo_output_file}"
					
					puts "Executing cmd: #{cmd}"
					system cmd
					File.open(algo_output_file){|file|
						while line = file.gets
							if line =~ /Result for ParamILS:/
								line = line.sub(/Result for ParamILS:/,"")
								puts "Result: #{line.strip}"
								solved, runtime, runlength, best_sol, seed = line.split(",").map!{|x|x.strip}
								instanceAndParamsHash[instance][param_string]["results"][seed.to_i] = [solved, runtime.to_f, runlength.to_i, best_sol.to_f, seed.to_i]
								break
							end
						end
						raise "No result in result file of #{cmd}:\n#{algo_output_file}" if instanceAndParamsHash[instance][param_string]["results"].empty?
					}
					File.delete(algo_output_file)
				end
			end
		end
		return true #results ready
	end

	#=== Collect runs I still have to do.
	algosToRun = []
	for instance in sorted_instances
		next unless instanceAndParamsHash.key?(instance) # If not all the original instances are in the instanceHash
		for param_string in instanceAndParamsHash[instance].keys
			params = instanceAndParamsHash[instance][param_string]["params"]
#			puts "getAlgoResultsForInstsAndParams algorun_config_id. Instance #{instance}, #{params.length} params #{params.keys.join(" ")}"
#			$stdout.flush

			algorun_config_id = get_algorun_config_id(algo, params, instance, instanceAndParamsHash[instance][param_string]["qual"], cutoff_time, cutoff_length)[0]
#			puts "getAlgoResultsForInstsAndParams after algorun_config_id"
#			$stdout.flush

			instanceAndParamsHash[instance][param_string]["algorun_config_id"] = algorun_config_id
			seeds = instanceAndParamsHash[instance][param_string]["seeds"]
			computedSeeds = computedSeeds(algorun_config_id)
			
			#=== Make sure not to use runs that were used before.
			neededSeeds = seeds - computedSeeds # Ruby set difference.
			unless neededSeeds.empty?
				for seed in neededSeeds 
					algosToRun << [algorun_config_id, seed]
				end
			end
			puts "Need #{neededSeeds.length} new runs. (Want #{seeds.length}, have #{computedSeeds.length}, #{(seeds&computedSeeds).length} matching) for algorun_config_id #{algorun_config_id}" 
		end
	end
	
	#=== Remove duplicates. (Can't use uniq! because array entries)
	uniqAlgosToRun = []
	conf_seed_hash = Hash.new
	for algoToRun in algosToRun
		conf, seed = algoToRun
		if conf_seed_hash.key?(conf) 
			next if conf_seed_hash[conf].include?(seed)
			conf_seed_hash[conf] << seed
		else
			conf_seed_hash[conf] = [seed]
		end
		uniqAlgosToRun << algoToRun
	end

	unless uniqAlgosToRun.empty?
		if oncluster==1 or oncluster=="1"
			runAlgosOnCluster(uniqAlgosToRun, false, algo)
		elsif oncluster==2 or oncluster=="2"
			runAlgosWithJobserver(uniqAlgosToRun)
		elsif oncluster==3 or oncluster=="3"
			runAlgosOnCluster(uniqAlgosToRun, true, "LS"+algo)
		else 
			runAlgosLocally(uniqAlgosToRun)
		end
	end
		
	#=== Read all results if they are ready at this point (i.e. if nothing has to be run using SGE).
	if algosToRun.length == 0 or not (oncluster=="1" or oncluster == 1)
		for instance in instanceAndParamsHash.keys
#			puts "Read run results for instance #{instance}" 
			for param_string in instanceAndParamsHash[instance].keys
#				puts "   Read run results for param_string #{param_string}" 
				algorun_config_id = instanceAndParamsHash[instance][param_string]["algorun_config_id"]
				seeds = instanceAndParamsHash[instance][param_string]["seeds"]

				results = readAlgoResults(algorun_config_id, seeds)
				unless results.class.to_s == "Array"
					raise "Results #{results} must be an array. Instance #{instance}, param_string #{param_string}" 
				end
				unless results[0].class.to_s == "Array"
					p results
					raise "results[0] #{results[0]} must be an array. Instance #{instance}, param_string #{param_string}" 
				end

				instanceAndParamsHash[instance][param_string]["results"] = Hash.new
				for seed in seeds
					instanceAndParamsHash[instance][param_string]["results"][seed] = results.shift
				end
			end
		end
		return true #results ready
	end
	return false #results not ready yet
end

=begin
###########################################################################
# Gets the results for running algo on multiple instances - keeps track of which runs are used and does not reuse them !
###########################################################################
def oldGetAlgoResultsForInstsAndParams(algo, instanceAndParamsHash, cutoff_time, cutoff_length, sorted_instances, oncluster=0, db=false)
	unless db
		#=== Do runs, save in instanceAndParamsHash, return
		for instance in instanceAndParamsHash.keys
			for param_string in instanceAndParamsHash[instance].keys
				params = instanceAndParamsHash[instance][param_string]["params"]
				qual = instanceAndParamsHash[instance][param_string]["qual"]
				reference = instanceAndParamsHash[instance][param_string]["reference"]
				minrun = instanceAndParamsHash[instance][param_string]["minrun"]
				maxrun = instanceAndParamsHash[instance][param_string]["maxrun"]

				paramstring = params.keys.map{|x| "#{x}=#{params[x]}"}.join(" ")
#TODO: make cutofftime dependent on reference_time
				cmd = "#{algo} #{instance} #{qual} #{cutoff_time} #{cutoff_length} #{minrun} #{maxrun} #{paramstring}"

				results = []
#				puts "Executing cmd: #{cmd}"
#				File.popen(cmd){|file| results = file.readlines} #=== Locally runs the algorithm.
				File.popen(cmd){|file|   #=== Locally runs the algorithm.
				#TODO: results = [lines]
					instanceAndParamsHash[instance][param_string]["result"] = []
					while line = file.gets
						puts line
						if line =~ /(.*), (#{float_regexp}), (-?\d+), (#{float_regexp}), (-?\d+)/
							res = $1
							runtime = $2
							runlength = $3
							quality_found = $4
							seed = $5
							instanceAndParamsHash[instance][param_string]["result"] << [res, runtime, runlength, quality_found, seed]
#						else
#							raise "Output line not matching the standards (SAT/UNSAT/TIMEOUT, float, integer, float, integer): #{line}"
						end
						raise "Problem while parsing for param_string #{param_string}" if line =~ /Problem with/
						raise "Exception in algorithm execution" if line =~ /Exception/ 
#						results = [[wall_clock_time, 0, 0]]
					end
				}
				raise "No line matching the result I want" if instanceAndParamsHash[instance][param_string]["result"] == []
			end
		end
		return true #results ready
	end

	#=== Collect runs I still have to do.
	algosToRun = []
	for instance in sorted_instances
		next unless instanceAndParamsHash.key?(instance) # If not all the original instances are in the instanceHash
		for param_string in instanceAndParamsHash[instance].keys
			params = instanceAndParamsHash[instance][param_string]["params"]
			algorun_config_id = get_algorun_config_id(algo, params, instance, instanceAndParamsHash[instance][param_string]["qual"], cutoff_time, cutoff_length)[0]
			instanceAndParamsHash[instance][param_string]["algorun_config_id"] = algorun_config_id
			numOfResults = numberOfAlgoResults(algorun_config_id)
			
			#=== Make sure not to use runs that were used before.
			maxrun = instanceAndParamsHash[instance][param_string]["maxrun"]
			if numOfResults < maxrun
				algosToRun << [algorun_config_id, maxrun-numOfResults]
			end
			puts "Need #{maxrun-numOfResults} new runs (want #{maxrun}, have #{numOfResults}) for algorun_config_id #{algorun_config_id}" 
		end
	end
	
	unless algosToRun.empty?
		if oncluster==1 or oncluster=="1"
			runAlgosOnCluster(algosToRun)
		elsif oncluster==2 or oncluster=="2"
			runAlgosWithJobserver(algosToRun)
		else 
			runAlgosLocally(algosToRun)
		end
	end
		
	#=== Read all results if they are ready at this point (i.e. if nothing has to be run using SGE).
	if algosToRun.length == 0 or not (oncluster=="1" or oncluster == 1)
		for instance in instanceAndParamsHash.keys
#			puts "Read run results for instance #{instance}" 
			for param_string in instanceAndParamsHash[instance].keys
#				puts "   Read run results for param_string #{param_string}" 
				algorun_config_id = instanceAndParamsHash[instance][param_string]["algorun_config_id"]

				minrun = instanceAndParamsHash[instance][param_string]["minrun"]
				maxrun = instanceAndParamsHash[instance][param_string]["maxrun"]
				result = readAlgoResults(algorun_config_id, minrun, maxrun)
				unless result.class.to_s == "Array"
					raise "Result #{result} must be an array. Instance #{instance}, param_string #{param_string}" 
				end
				unless result[0].class.to_s == "Array"
					p result
					raise "result[0] #{result[0]} must be an array. Instance #{instance}, param_string #{param_string}" 
				end

				instanceAndParamsHash[instance][param_string]["result"] = result
			end
		end
		return true #results ready
	end
	return false #results not ready yet
end
=end

###########################################################################
# The new and improved version of running algorithms: we get a list of entry_param pairs, run them all (as an instanceAndParamsHash) and directly modify the entries.
###########################################################################
def getAlgoResultsForEntryParamsPairs(algo, entry_params_pairs, cutoff_time, cutoff_length, oncluster=0, db=true)
	$total_cputime = 0 unless $total_cputime
	#=== Check which of the entry param pairs we actually need to run.
#	puts "Collect runs I need to do."
	entry_params_pairs_to_run = []
	count = 1
	for entry_params in entry_params_pairs
#		p entry_params 
		puts "count #{count}/#{entry_params_pairs.length}" if count.divmod(100)[1]==0
		count += 1
		entry, params = entry_params
#		p entry
#		p params 
		entry_params_pairs_to_run << entry_params unless entry["resultForState"].key?(state_string(params))
	end
	return true if entry_params_pairs_to_run.empty?

#	puts "Construct instanceAndParamsHash for #{entry_params_pairs_to_run.length} entry_params pairs."
	#=== Construct instanceAndParamsHash to run.
	instanceAndParamsHash = Hash.new

	for entry_params in entry_params_pairs_to_run
		entry, params = entry_params
		param_string = state_string(params)
		inst = entry["name"]
		instanceAndParamsHash[inst] = Hash.new unless instanceAndParamsHash.key?(inst)
		
		unless instanceAndParamsHash[inst].key?(param_string)
			instanceAndParamsHash[inst][param_string] = {"qual" => entry["qual"], "reference"=>entry["reference"], "params" => params, "seeds"=>[]}
		end
		instanceAndParamsHash[inst][param_string]["seeds"] << entry["seed"]
	end

#	puts "getAlgoResultsForInstsAndParams."

	#=== Run everything in the instanceAndParamsHash.
	results_ready = getAlgoResultsForInstsAndParams(algo, instanceAndParamsHash, cutoff_time, cutoff_length, instanceAndParamsHash.keys, oncluster, db)

	if results_ready
		#=== Retrieve those results we had to run.
		for entry_params in entry_params_pairs_to_run
			entry, params = entry_params
			param_string = state_string(params)
			inst = entry ["name"]
			seed = entry["seed"]
			
			result = instanceAndParamsHash[inst][param_string]["results"][seed]
			unless result
				puts "ERROR - empty result."
				p instanceAndParamsHash[inst]
				p instanceAndParamsHash[inst][param_string]
				p instanceAndParamsHash[inst][param_string]["results"]

				raise "Empty result for instance #{inst}, seed #{seed}, and param_string #{param_string}"
			end
			entry["resultForState"][param_string] = result
			$total_cputime += result[1].to_f
		end
	end
	return results_ready
end



#=== Run numbers for each instances MUST BE INCREASING !
def makeStandardInstanceHash(array_of_detailed_instances, state_as_string)
	instanceHash = Hash.new
	for entry in array_of_detailed_instances
		inst = entry["name"]
		run = entry["run"]
		instanceHash[inst] = {"runs"=>0, "desired_qual"=>entry["qual"], "reference"=>entry["reference"], "result"=>[]} unless instanceHash.key?(inst)
		instanceHash[inst]["runs"] += 1
		instanceHash[inst]["result"] << entry["resultForState"][state_as_string]
	end
	return instanceHash
end


###########################################################################
# Returns the set of instances and their desired solution qualities contained in a file.
###########################################################################
def getInstanceHash(instanceset_name, algo="none")
	#=== Special case: the algorithm does not take input instances.
	return [{"_#{algo}_dummyInstance"=>{"desired_qual"=> 0, "reference"=>0}}, ["_#{algo}_dummyInstance"]] if instanceset_name == "_"

	#=== Read set of instances to run on. (array of arrays [instance_name, opt])
	instancesSorted = []
	instanceHash = Hash.new
	File.open(instanceset_name){|file| 
	    while line=file.gets;
		raise "Incorrect number of entries #{line.strip.split.length} in #{line}" unless line.strip.split.length >= 2  
		entries = line.chomp.strip.split
		inst = entries[0]
		qual = entries[1]
		reference = entries[2]
		instanceHash[inst] = Hash.new
		instanceHash[inst]["desired_qual"] = qual
		instanceHash[inst]["reference"] = reference
		instancesSorted << inst
	    end
	}
	return [instanceHash, instancesSorted]
end

###########################################################################
# Returns the results for all runs on each instance in an instanceset_file: [inst_name, opt_sol, [resRun1, resRun2]], where resRuni is [time, steps, foundsolqual]
###########################################################################
def getResultsForInstanceFile(algo, params, cutoff_time, cutoff_length, minseed, maxseed, instanceset_name, state_as_string, oncluster=0, db=false)
	$total_cputime = 0 unless $total_cputime
	instanceHash, instancesSorted = getInstanceHash(instanceset_name, algo)
	
	instanceAndParamsHash = Hash.new
		
	seeds = []
	for seed in minseed..maxseed
		seeds << seed
	end 
		
	for instance in instancesSorted
		instanceAndParamsHash[instance] = Hash.new
		instanceAndParamsHash[instance][state_as_string] = {"qual"=>instanceHash[instance]["desired_qual"], "reference"=>instanceHash[instance]["reference"], "seeds"=>seeds, "params" => params}
	end

	results_ready = getAlgoResultsForInstsAndParams(algo, instanceAndParamsHash, cutoff_time, cutoff_length, instanceAndParamsHash.keys, oncluster, db)

	if results_ready
		entries = []
		instances = []
		#=== Retrieve those results we had to run.
		for instance in instancesSorted
			for seed in minseed ... maxseed+1
				entry = {"name"=>instance, "qual"=>instanceHash[instance]["desired_qual"], "seed" => seed, "reference"=>instanceHash[instance]["reference"], "resultForState"=>Hash.new}
				result = instanceAndParamsHash[instance][state_as_string]["results"][seed]
				$total_cputime += result[1].to_f
				entry["resultForState"][state_as_string] = result
				entries << entry
			end
		end
		return [entries, results_ready]
	end
end