require "param_reader.rb"
require "global_helper.rb" # for instance hash

# =========================================
#                  put string to output and outfile
# =========================================
def output(s)
	if @output_level >= 50
		puts s
		@out.puts s
		@out.flush
	end
end


# =========================================
# Get quality of a parameter setting without extra runs.
# =========================================
def eval(state)
	state_as_string = state_string(state)
	if @approach =~ /basic/ or @approach == "random"
		return normal_eval(strip_state(state),state_as_string)
	else
		return partial_eval(strip_state(state),state_as_string)
	end
end


# =========================================
# Output evaluation of a state
# =========================================
def str(state)
	val, num_samples = eval(state)
	return "#{state_string(state)} (%f [based on a total of %d runs])" % [val, num_samples] if val 
	return "#{state_string(state)} (-- [no runs])"
end


# =========================================
# Output more detailed information for a state.
# =========================================
def output_result(state)
#	for test_inst in @test_instance_names
	stripped_state = strip_state(state)
	state_as_string = state_string(stripped_state)

	entriesForObjective = []
	instance_names = []
	for entry in @instances
		if entry["resultForState"].key?(state_as_string)
			entriesForObjective << entry 
			instance_names << entry["name"] unless instance_names.include?(entry["name"])
		else
			break
		end
	end
	raise "FATAL! Number of results #{numEvaluated(state)} does not equal results from the beginning #{entriesForObjective.length}" if (@approach=~ /focused/ and not numEvaluated(state)==entriesForObjective.length)
	
	objectives = getAllObjectives(@algo, @run_obj, entriesForObjective, instance_names, state_as_string, @cutoff_time, @cutoff_length)
#	objectives = getObjectivesForComputedInstances(@algo, @run_obj, entriesForObjective, @instances_sorted, state_as_string)
	#instanceHash = makeStandardInstanceHash(@instances, state_string(stripped_state))

	outputObjectives(instance_names, objectives)
end


# =========================================
# Adaptive neighbourhood relation that takes into account conditional parameters.
# =========================================
def neighbourhood(current_state)
	state = strip_state(current_state)
	result = []
	for param in state.keys
		next if @fixed_ass.key?(param)
		for value in @domain[param]
			unless current_state[param] == value
				new_state = current_state.dup
				new_state[param] = value

				next if forbidden(new_state, @forbidden_combos)
				result << new_state
			end
		end
	end
	if result == []
		puts "fixed_ass:"
		p @fixed_ass
		puts "state.keys"
		p state.keys
		raise "neighbourhood can't be empty!!"
	end
	return result
end


# =========================================================================================
# Output incumbent state is found.
# =========================================================================================
def output_incumbent_state(state)
	val, numSamples = eval(state)
	val = 100000000 if val == nil

	#=== Fill in the unnecessary parameters with their default values for output.
	filled_state = state.dup
	for param in @params
		if state.key?(param)
			filled_state[param] = state[param]
		else
			filled_state[param] = @default[param]
			filled_state[param] = @fixed_ass[param] if @fixed_ass.key?(param)
		end
	end

	numeval = @evaluation_count

	@param_ils_traj_file.print "#{$total_cputime}, #{val}, #{numeval}, #{@iteration}, #{@numFlip}, "
	@param_ils_traj_file.puts @params.map{|x|filled_state[x]}.join(", ") # output the current values
	@param_ils_traj_file.flush

	puts "Incumbent: #{$total_cputime}, #{val} [#{numSamples}]. With state " + @params.map{|x| "#{x}=#{filled_state[x]}"}.join(", ") 
	@out.puts "Incumbent: #{$total_cputime}, #{val} [#{numSamples}]. With state " + @params.map{|x| "#{x}=#{filled_state[x]}"}.join(", ") 
	@out.flush
end


# =========================================================================================
# Set a new incumbent state.
# =========================================================================================
def setNewIncumbent(state)
	@incumbent_state = state.dup
	output_incumbent_state(state)
end

# =========================================================================================
# Return how many entries have been ran with a parameter setting.
# =========================================================================================
def numEvaluated(state)
	state_as_string = state_string(state)
	return 0 unless @cachedRunObjectives.key?(state_as_string)
	return @cachedRunObjectives[state_as_string].length
end


# =========================================================================================
# Get the results for instance-seed combinations for a given param config.
# =========================================================================================
def getRunsOnEntriesWithParams(entries, param_state, test=false)
	raise "state_string error: strip_state(strip_state(s)) != strip_state(s): \n#{state_string(strip_state(param_state))} != \n#{state_string(param_state)}" unless state_string(param_state) == state_string(strip_state(param_state))
	stripped_state = strip_state(param_state)
	state_as_string = state_string(param_state)

	#=== Assert that we don't ask for entries we already know. 
	for entry in entries
		if entry["resultForState"][state_as_string]
			puts "Non-Empty result! " 
			puts "State"
			p state_as_string
			puts "Entry"
			p entry
			p entry["resultForState"]
			p entry["resultForState"][state_as_string]
			raise "Non-Empty result in getRunsOnEntriesWithParams before asking for it."
		end
	end

	getAlgoResultsForEntryParamsPairs(@algo, entries.map{|entry| [entry, stripped_state]}, @cutoff_time, @cutoff_length, @oncluster, @db)

	#=== Assert that we now know all the entries. 
	for entry in entries
		unless entry["resultForState"][state_as_string]
			puts "Empty result! " 
			puts "State"
			p state_as_string
			puts "Entry"
			p entry
			p entry["resultForState"]
			p entry["resultForState"][state_as_string]
			raise "Empty result"
		end
	end

	#=== Compute the objective for each entry.
	results = entries.map{|entry| singleRunObjective(@algo, @run_obj, entry["resultForState"][state_as_string], entry["qual"], entry["reference"], @cutoff_time, @cutoff_length)}
	
	unless test
		@evaluation_count += results.length
		if $total_cputime > @lastOutputCPUTime + 10
			@lastOutputCPUTime  = $total_cputime
			output("#{@evaluation_count}/#{@maxEvals}, #{$total_cputime}/#{@tunerTimeout}")
		end
		#=== Keep track of results for each <param, value> pair.
		for entry in entries
			entry["scores"] = Hash.new unless entry.key?("scores")
			for param in stripped_state.keys
				entry["scores"][param] = Hash.new unless entry["scores"].key?(param)
				value = stripped_state[param]
				entry["scores"][param][value] = [] unless entry["scores"][param][value]
				entry["scores"][param][value] << singleRunObjective(@algo, @run_obj, entry["resultForState"][state_as_string], entry["qual"], entry["reference"], @cutoff_time, @cutoff_length)
			end
		end
	end
	
	return results
end


# =========================================================================================
# Add new runs for param_state.
# Invariants: Before and after this, @incumbent_state is one of the states with the most evaluations, and amongst those, it is the best.
# The number of runs is limited by @N.
# =========================================================================================
def addNewRuns(param_state, numNewRuns)
#	puts "addNewRuns begin"
	$stdout.flush	
	#=== Set up short identifiers.
	state_as_string = state_string(param_state)
	@cachedRunObjectives[state_as_string] = [] unless @cachedRunObjectives.key?(state_as_string)
	res = @cachedRunObjectives[state_as_string]

	incumbent_state_string = state_string(@incumbent_state)
	@cachedRunObjectives[incumbent_state_string] = [] unless @cachedRunObjectives.key?(incumbent_state_string)
	inc_res = @cachedRunObjectives[incumbent_state_string]

	#=== Only get results up to maximum of @N.
	return if res.length >= @N
	if res.length + numNewRuns > @N
		res.concat getRunsOnEntriesWithParams(@instances[res.length...@N], param_state)
	else
		res.concat getRunsOnEntriesWithParams(@instances[res.length...res.length+numNewRuns], param_state)
	end
		
	#=== If the state is the incumbent one, output its new value.
	if state_as_string == incumbent_state_string
		setNewIncumbent(param_state)
		output "With same incumbent state as before: #{str(param_state)}"
		return
	end
	
	diffInRuns = res.length - inc_res.length
	
	#=== Assert that param_state has only a maximum of numNewRuns runs more than the incumbent one.
	raise "Can only have #{numNewRuns} more run than best, but have #{diffInRuns} more for state #{str(param_state)} than for incumbent #{str(@incumbent_state)}" if diffInRuns > numNewRuns

	#=== If there are more runs for param_state than for incumbent, catch up.
	if diffInRuns > 0
		output "State #{str(param_state)} got #{diffInRuns} more run than incumbent #{str(@incumbent_state)}, catching up. (numEvalNew=#{res.length}, numEvalInc=#{inc_res.length})"
		addNewRuns(@incumbent_state, diffInRuns)
		
		raise "numEvaluated should now match, but doesn't: #{numEvaluated(param_state)} != #{numEvaluated(@incumbent_state)}" unless numEvaluated(param_state) == numEvaluated(@incumbent_state)
	end
	
	#=== If there are equally many runs for param_state and incumbent, compare the two and change incumbent if the other one is better.
	if res.length == inc_res.length
		diff_of_objs = combinationOfObjectiveFunctions(@algo, @overall_obj, res, @run_obj, @cutoff_time, @cutoff_length) - combinationOfObjectiveFunctions(@algo, @overall_obj, inc_res, @run_obj, @cutoff_time, @cutoff_length)
#		puts "diff_of_objs = #{diff_of_objs}, old incumbent res: #{inc_res.join(" ")}, new one #{res.join(" ")}"
		setNewIncumbent(param_state) if diff_of_objs.to_f < @eps
	end
	
	return
end



# =========================================================================================
# Helper function for better. 
# Returns true iff combination_of_results_1 - combination_of_results_2 <= eps, 
# i.e. iff the first state is better or equal to the second one using the lesser number of samples.
# =========================================================================================
def isBetterWithLessEvals(state1, state2)
#	puts "comparing \n#{state1} and \n#{state2}"
	numEval1 = numEvaluated(state1)
	numEval2 = numEvaluated(state2)
	minNumEval = [numEval1, numEval2].min
#	puts "With #{minNumEval} evals"
	
	res1 = @cachedRunObjectives[state_string(state1)]
	res2 = @cachedRunObjectives[state_string(state2)]
	
	result = ((combinationOfObjectiveFunctions(@algo, @overall_obj, res1[0...minNumEval], @run_obj, @cutoff_time, @cutoff_length).to_f - combinationOfObjectiveFunctions(@algo, @overall_obj, res2[0...minNumEval], @run_obj, @cutoff_time, @cutoff_length)).to_f <= @eps)
#	puts result
	return result
end

# =========================================================================================
# State1 dominates state2 iff it has at least as many samples and is better with the lesser amount of samples.
# =========================================================================================
def dominates(state1, state2)
	return ((numEvaluated(state1) >= numEvaluated(state2)) and (isBetterWithLessEvals(state1, state2)) )
end

# =========================================================================================
# Better: this is a completely symmetric comparison (except in the case of ties, in which case true is returned).
# At the end of this, one state must have at least as many evals as the other one and outperform it with the lower number of evals of the two.
# Cases:
# 1) state1 has more or equal number of evals and is better with those: return false (first is better)
# 2) state2 has more or equal number of evals and is better with less evals: return true (second is better)
# 3) state1 has less evals but is better with those: execute runs of state1 until either 1) or 2)
# 4) state2 has less evals but is better with those: execute runs of state2 until either 1) or 2)
# =========================================================================================
def better(new_state, old_state)
	if @approach =~ /basic/
		newVal, numSamples = normal_eval(new_state, state_string(new_state))
		oldVal, numSamples = normal_eval(old_state, state_string(old_state))
		return newVal <= oldVal
	elsif @approach =~ /focused/
		#=== Everybody gets at least one new evaluation.
		if numEvaluated(new_state) < numEvaluated(old_state)
			addNewRuns(new_state, 1)
			@numEvaluationsThisIteration += 1
		elsif numEvaluated(new_state) > numEvaluated(old_state)
			addNewRuns(old_state, 1)
			@numEvaluationsThisIteration += 1
		else
			addNewRuns(new_state, 1) 
			@numEvaluationsThisIteration += 1

			if numEvaluated(new_state) > numEvaluated(old_state) # not always the case because if old_state is incumbent then addNewRun triggers a run of it.
				addNewRuns(old_state, 1) 
				@numEvaluationsThisIteration += 1
			end
		end

		#=== If everything is clear, return.
		return true if dominates(new_state, old_state)
		return false if dominates(old_state, new_state)
		
		#=== Otherwise, the one with less evaluations looks better.
		#Case 1: new one has less and is better.
		if numEvaluated(new_state) < numEvaluated(old_state)
			numNewRuns=1
			while not (dominates(new_state, old_state) or dominates(old_state, new_state))
#				numNewRuns = [numNewRuns, numEvaluated(old_state)-numEvaluated(new_state)].min
				addNewRuns(new_state, numNewRuns)
	#			@numEvaluationsThisIteration += numNewRuns
#				numNewRuns *= 2
			end
		end

		#Case 2: old one has less and is better.
		if numEvaluated(old_state) < numEvaluated(new_state)
			numNewRuns=1
			while not (dominates(new_state, old_state) or dominates(old_state, new_state))
#				numNewRuns = [numNewRuns, numEvaluated(new_state)-numEvaluated(old_state)].min
				addNewRuns(old_state, numNewRuns)
	#			@numEvaluationsThisIteration += newRuns
#				numNewRuns *= 2
			end
		end
		
#		puts "end better"
#		$stdout.flush

		return true if dominates(new_state, old_state)
		return false if dominates(old_state, new_state)
		raise "Still undecided at end of better between new state #{str(new_state)} and old state #{str(old_state)}. This shouldn't be possible."
	else raise "Unknown approach: #{@approach}"
	end
end

# =========================================
# Evaluate a state in the good old fashion: get an objective function value for each instance from the existing runs and then combine them.
# =========================================
def normal_eval(stripped_state, state_as_string)
	return @cachedObjectives[state_as_string] if @cachedObjectives.key?(state_as_string)

	getAlgoResultsForEntryParamsPairs(@algo, @instances.map{|entry| [entry, stripped_state]}, @cutoff_time, @cutoff_length, @oncluster, @db)
	obj = getGlobalObjective(@algo, @run_obj, @overall_obj, @instances, state_as_string, @cutoff_time, @cutoff_length)

	@cachedObjectives[state_as_string] = obj
	@evaluation_count += @N
	
	#=== Have evaluated a new state, check if it is the new incumbent.
	new_obj, numEval = obj
	old_obj, numEval = @cachedObjectives[state_string(@incumbent_state)]
	diff_of_objs = new_obj - old_obj
	setNewIncumbent(stripped_state) if diff_of_objs < @eps

	if $total_cputime > @lastOutputCPUTime + 10
		@lastOutputCPUTime  = $total_cputime
		output("#{@evaluation_count}/#{@maxEvals}, #{$total_cputime}/#{@tunerTimeout}")
	end

	#=== Keep track of results for each <param, value> pair.
	@scores = Hash.new unless @scores
	for param in stripped_state.keys
		@scores[param] = Hash.new unless @scores.key?(param)
		value = stripped_state[param]
		@scores[param][value] = [] unless @scores[param].key?(value)
		@scores[param][value] << new_obj
	end
#	p @scores
	return obj
end

# =========================================
# Partially evaluate a state - i.e. only use the runs that are available.
# =========================================
def partial_eval(stripped_state, state_as_string)
	entriesForObjective = []
	for entry in @instances
		if entry["resultForState"].key?(state_as_string)
			entriesForObjective << entry 
		else 
			break
		end
	end

	return getGlobalObjective(@algo, @run_obj, @overall_obj, entriesForObjective, state_as_string, @cutoff_time, @cutoff_length)
end

# =========================================
# For learning component, compute promise of parameter instantiations.
# =========================================
=begin
# Learning component under development.
def promiseOfFlippingParamFromTo(param, currentValue, targetValue)
	output "promise of #{param}: #{currentValue}->#{targetValue}"
	if @approach =~ /basic/
		#=== Unseen parameters and values have the same promise: average of the rest.
		return nil unless @scores
		return nil unless @scores[param] 
		return nil unless @scores[param][targetValue]
		return nil unless @scores[param][currentValue]
		
		objsWithCurrVal = @scores[param][currentValue]
		objsWithTargetVal = @scores[param][targetValue]
	else
		#=== TODO: implement normalization per instance.
		objsWithCurrVal = []
		objsWithTargetVal = []
		for entry in @instances
			if entry.key?("scores") and entry["scores"].key?(param) and entry["scores"][param].key?(currentValue)
				objsWithCurrVal += entry["scores"][param][currentValue]
			end
			if entry.key?("scores") and entry["scores"].key?(param) and entry["scores"][param].key?(targetValue)
				objsWithTargetVal += entry["scores"][param][targetValue]
			end
		end
		return nil if objsWithCurrVal.empty?
		return nil if objsWithTargetVal.empty?
	end
	#=== High objective function values => low promise.
	require "stats_ils.rb"
	p objsWithCurrVal
	avg_current = objsWithCurrVal.quantile(0.25) # Make weighted average later.
	output avg_current
	avg_target = objsWithTargetVal.quantile(0.25) # Make weighted average later.
	output avg_target
	avg_promise = avg_current - avg_target
	return avg_promise
end
=end


# =========================================
# Random search, for comparison.
# =========================================
def random_search()
	output "========================================================\nStarting RANDOM SEARCH \n========================================================"
	count = 0
	current_state = init_random()
	setNewIncumbent(current_state)
	visited_states = Hash.new
	visited_states[state_string(current_state)] = 1
	
	while count < @M
		while true
			random_state = init_random()
			next if visited_states.key?(state_string(random_state))
			visited_states[state_string(random_state)] = 1
			break
		end
				
		if better(random_state, @incumbent_state, "normal")
#			checkIfNewIncumbent(random_state)
			output("          -> Take improving step to random #{str(random_state)}\n\n")  
		else 
#			checkIfNewIncumbent(current_state)
			output "        -> Worse random: #{str(random_state)}"
		end
		count += 1
	end
end

# =========================================================================================
# =========================================================================================
#                  HERE BEGINS THE ITERATED LOCAL SEARCH AND ITS SUBFUNCTIONS
# =========================================================================================
# =========================================================================================


# =========================================
# Stop search if one of the termination criteria is fulfilled.
# =========================================
def haveToStop()
	return true if @evaluation_count>=@maxEvals
	return true if $total_cputime >= @tunerTimeout
	return true if @iteration > @maxIts
	return false
end

# =========================================
# Standard ILS.
# =========================================
def iterated_local_search()
	#=== Init.
	if @init_def
		current_state = init_default()
	else
		current_state = init_random()
	end
#	p current_state
#	p strip_state(current_state)
	setNewIncumbent(current_state)

	#=== Perform @R random steps.
	@numEvaluationsThisIteration = 0
	for i in 0...@R
		random_state = init_random()
		
		if better(random_state, current_state)
			output("          -> Take improving step to random #{str(random_state)}\n\n")  
			current_state = random_state
		else 
			output "        -> Worse random: #{str(random_state)}"
		end
	end

#	current_state = init_random()
	@numEvaluationsThisIteration=0
#	last_improving_iteration = 1
	last_ils_state = basic_local_search(current_state)
	current_state = last_ils_state.dup
  
	@iteration += 1
	while not haveToStop()
		output("#{@evaluation_count}/#{@maxEvals}, #{$total_cputime}/#{@tunerTimeout}")

                puts "iteration #{@iteration}, flip #{@numFlip}, evaluation count #{@evaluation_count}"
                @out.puts "iteration #{@iteration}, flip #{@numFlip}, evaluation count #{@evaluation_count}"

		current_state = perturbation(current_state, @pertubation_strength)
		current_state = basic_local_search(current_state)
		
		acc_state = acceptance_criterion(last_ils_state, current_state)
		#=== With low probability, random restart.
		if rand(100) < @p_restart*100
			acc_state = init_random() 
			output "Random re-initialisation to #{str(acc_state)}"
		end

		current_state = acc_state
		last_ils_state = current_state
		
		@iteration += 1
	end
end

# =========================================
# Init local search with default.
# =========================================
def init_default()
	init_state = Hash.new
	for param in @params
		init_state[param] = @default[param]
	end
	for key in @start_ass.keys
		init_state[key] = @start_ass[key]
	end
	for key in @fixed_ass.keys
		init_state[key] = @fixed_ass[key]
	end
	return init_state
end

# =========================================
# Init local search with random values whose combination is allowed.
# =========================================
def init_random()
	is_forbidden = true
	while is_forbidden
		init_state = Hash.new
		for param in @params
			init_state[param] = @domain[param][rand(@domain[param].length)]
		end
		for key in @fixed_ass.keys
			init_state[key] = @fixed_ass[key]
		end
		is_forbidden = forbidden(init_state, @forbidden_combos)
	end
	return init_state
end

# =========================================
# Basic local search.
# =========================================
def basic_local_search(start_state)
	current_state = start_state

	@numEvaluationsThisIteration = 0
	output "   BLS in iteration #{@iteration}, start with #{str(current_state)}\n"
	visited_states = [state_string(current_state)]
	changed = true
	while changed and not haveToStop() # local search.
		changed = false
		neighbours = neighbourhood(current_state)
		#=== Move to the best neighbour
		while neighbours.length > 0 and not haveToStop()
				if @approach =~ /learning/
=begin
				# Learning component under development.
				scores = []
				minscore = 1e10
				maxscore = -1e10
				
				#=== Determine which param is different in neighbour and then check how promising it is to change that param.
				output "First scores"
				for i in 0...neighbours.length
					neighb = neighbours[i]
					diff_param = nil
					for param in current_state.keys
						if current_state[param] != neighb[param]
							raise "Neighbour differs in more than one param: #{diff_param} and #{param}" if diff_param
							diff_param = param
						end
					end
					raise "Neighbour doesn't differ in any param: #{current_state}\n#{neighb}" unless diff_param
					score = promiseOfFlippingParamFromTo(diff_param, current_state[diff_param], neighb[diff_param])
					output neighb
					output score
					scores << score
					minscore = [score, minscore].min if score
					maxscore = [score, maxscore].max if score
				end
				
				#=== Remove all negative scores (set 0), then fill in nils with avgscore.
				sumscores = 0
				for i in 0...scores.length
					if scores[i]
						scores[i] = 0 if scores[i] < 0
						sumscores += scores[i]
					end
				end
				avgscore = (sumscores+0.0)/scores.length
				sumscores = 0
				for i in 0...scores.length
					scores[i] = avgscore unless scores[i]
					sumscores += scores[i]
				end

				#=== Normalize.
				if sumscores == 0
					for i in 0...scores.length
						scores[i] = 1
					end
					sumscores = scores.length + 0.0
				end
				scores.map!{|score| score /= sumscores}
				output "Before uniform mass added"
				output scores

				#=== Smooth by adding some uniform probability mass everywhere.
				alpha = 0.2
				sumscores = 0
				unif_mass = 1.0/scores.length
				for i in 0...scores.length
					scores[i] = alpha * unif_mass + (1-alpha) * scores[i]
					sumscores += scores[i]
				end
				scores.map!{|score| score /= sumscores}

				output "Final scores"
				output scores

				#=== Pick one neighbour with probability proportional to the scores.
				rand_num = rand
				sum = 0
				for i in 0...neighbours.length
					sum += scores[i]
					break if sum > rand_num
				end
				unless sum > rand_num
					p scores
					p i
					raise "sum of scores #{sum} not bigger than rand_num #{rand_num} at end of loop" 
				end
				rand_index = i
				output "Picking neighbour #{rand_index} with normalized score #{scores[rand_index]} with u=#{rand_num} (out of #{neighbours.length} neighbours)"
=end
			else
				rand_index = rand(neighbours.length)
			end
			neighbour = neighbours[rand_index]
			neighbours.delete_at(rand_index)
#			puts state_string(neighbour) 
			#=== To avoid loops, disallow state we have been at in this iteration.
			next if visited_states.include?(state_string(neighbour)) 
			visited_states << state_string(neighbour) 
#			puts "Looking at it."

			output("    Changing #{changed_between_states(current_state, neighbour)}, evaluating ...")
			if better(neighbour, current_state)
				output("          -> Take improving step to neighbour #{str(neighbour)} with flip #{@numFlip}\n\n")  
				current_state = neighbour
				#=== Bonus runs in FocusedILS.
				if @approach =~ /focused/
					@numEvaluationsThisIteration = @maxbonus if @numEvaluationsThisIteration > @maxbonus 
					addNewRuns(current_state, @numEvaluationsThisIteration)
					output("          -> After #{@numEvaluationsThisIteration} bonus runs: #{str(current_state)}\n\n")  
					@numEvaluationsThisIteration=0
				end
				changed = true
				break
			else 
				output "        -> worse: #{str(neighbour)}"
			end
		end
		@numFlip += 1
	end
	if @approach =~ /focused/
		@numEvaluationsThisIteration = @maxbonus if @numEvaluationsThisIteration > @maxbonus 
		addNewRuns(current_state, @numEvaluationsThisIteration )
		output("          -> After #{@numEvaluationsThisIteration} bonus runs for LM: #{str(current_state)}\n\n")  
		@numEvaluationsThisIteration=0
	end
	
	output "   LM for iteration #{@iteration}: #{str(current_state)}\n"
#	puts "   LM for iteration #{@iteration}:  #{str(current_state)}\n"
#	@out.puts "   LM for iteration #{@iteration}: #{str(current_state)}\n"
	output "\n========== DETAILED RESULTS (iteration #{@iteration}): =========="
	output_result(current_state)
	output "================================================\n"
	return current_state
end

# =========================================
# Simple pertubation.
# =========================================
def perturbation(state, strength)
	result_state = state.dup
	strength.times{|x|
		neighbours = neighbourhood(result_state)
		rand_index = rand(neighbours.length)
		result_state = neighbours[rand_index]
		output "    perturb to ---> #{state_string(result_state)}" # Don't need to evaluate at this point!
	}
	return result_state
end

# =========================================
# Acceptance criterion BETTER.
# =========================================
def acceptance_criterion(last_ils_state, current_state)
	if state_string(last_ils_state) == state_string(current_state)
		output "same state as last ILS: #{state_string(last_ils_state)}"
		@accepted_last = "same"
		return current_state
	end

	if better(current_state, last_ils_state)  
		output "   Accepting new better local optimum: #{str(current_state)}"
		@accepted_last = "acc"
		return current_state
	end

	output "rejecting worse #{str(current_state)}, going back to #{str(last_ils_state)}"
	@accepted_last = "rej"
	return last_ils_state
end

# =========================================
# Return what changed between two states.
# =========================================
def changed_between_states(old_state, new_state)
	res = []
	for key in old_state.keys
		res << "#{key}: #{old_state[key]}->#{new_state[key]}" if old_state[key] != new_state[key]
	end
	return res
end

def output_help(out)
	out.puts "\n======================================================================================================"
	out.puts "PARAM_ILS version 1.0, Frank Hutter, 2007. Optimizes parameters of an algorithm on a set of instances."
	out.puts "For more details, see: \"Frank Hutter, Holger Hoos, and Thomas Stueztle:"
	out.puts  "Automatic Algorithm Configuration based on Local Search. In Proc. of AAAI-07\"."
	out.puts "Usage: ruby param_ils_1_0_run.rb, followed by the following optional arguments (defaults in [parentheses])"

	out.puts "General"
	out.puts "-------\n"
	out.puts "-algo <X>, where <X> calls (a wrapper of) the algorithm to be optimized. [#{@algo}]"
	out.puts "           This wrapper/algorithm has to follow the input/output format in README.TXT" 
	out.puts "-run_obj <X>, where <X> in {runtime, runlength, speedup} [#{@run_obj}] (or user-defined in file algo_specifics.rb)"
	out.puts "-overall_obj <X>, where <X> in {mean, adj_mean, median, q90, geomean} [#{@overall_obj}]\n"
	out.puts "-cutoff_time <X>, where <X> is a float [#{@cutoff_time}]"
	out.puts "-cutoff_length <X>, where <X> is an integer [#{@cutoff_length}]"
	out.puts "-N <X>, where X is an integer indicating the maximal number of samples obtained per config [#{@N}]."# (BasicILS(N) or FocusedILS with a bound on each N(theta)) [#{@N}]"
	out.puts "-numRun <X>, where X is an integer indicating the how manieth run of this tuning scenario this is[#{@numRun}]."
	out.puts "-instance_seed_file <X>, where X is a filename [#{@instance_seed_file}]."
	out.puts "-test-instance_seed_file <X>, where X is either empty or a filename. Either means use half the training set as validation. [#{@test_instance_seed_file}]."
	out.puts "-test <X>, where X is IN {0,1}. In combination with scenario, use test set instead of validation set."

	out.puts "\nWHEN SHOULD the tuner TERMINATE?"
	out.puts "--------------------------------"
	out.puts "-maxEvals <X>, where <X> is an integer defining the budget of single algorithm evaluations [#{@maxEvals}]"
	out.puts "-tunerTimeout <X>, where <X> is a float (seconds) [#{@tunerTimeout}]"
	
	out.puts "\nGeneral CONTROL:"
	out.puts "----------"
	out.puts "-fixed <X>, where <X> = \"<param1> <val1> <param2> <val2> ...\" and those values will be fixed throughout the tuning [\"#{@fix_input}\"]"
	out.puts "-output_level <X>, where <X> is an integer (higher = more output) [#{@output_level}]"

	out.puts "-db <X>, where <X> in {0,1} [#{@db}]. I use a MQSQL database to store algorithm runs." 
#	out.puts "         so I don't have to redo them if I redo an experiment. You are welcome to look "
#	out.puts "         at my scripts and adapt them to fit your purpose, but I will not support this feature."
#	out.puts "         When DB is used, <algo> means the name of the algorithm in the DB."
	
	out.puts "-jobserver <X>, where <X> in {0,1} [#{@jobserver}]. For big jobs, I distribute calls to several machines."
#	out.puts "              using jobserver.rb, written by Christian Bang. I will not support this feature, but "
#	out.puts "              you are welcome to look at the scripts I use to dispatch the jobs"

	out.puts "\nCONTROL OF PARAM_ILS:"
	out.puts "---------------------"
	out.puts "-approach <X>, where <X> in {random, basic, focused, basiclearning, focusedlearning} [#{@approach}]"
	out.puts "-init <X>, where <X> is either in {0,1} (0=random, 1=default), or <X> = \"<param1> <val1> <param2> <val2> ...\" [#{@init_def}]"

	out.puts "\nExamples:\n"
#	out.puts "ruby ../scripts/param_ils_1_0_run.rb gls4mpe -run_obj approx -overall_obj mean -N 50 -numParamILSRuns 10 -maxEvals 20000 -instanceFile gridGH25GW25K2-noquals-train-1-to-50.txt -db 1 -cluster 2"
#	out.puts "ruby ../scripts/explore_new_focused_ils.rb saps median_runlength adj_mean 100 10 2000000 normal 10 200 sw-sat-saps-1M-to-2M-train-first100.txt params.txt 1 0 0 2"
#	out.puts "ruby ../scripts/explore_new_focused_ils.rb saps median_runlength adj_mean 100 10 2000000 random 10 200 sw-sat-saps-1M-to-2M-train-first100.txt params.txt 1 0 0 2"
#	out.puts "ruby ../scripts/param_ils_1_0_run.rb -algo saps -db 1 -cluster 2"

#	out.puts "ruby ../scripts/explore_new_focused_ils.rb sat4j median_runlength adj_mean 100 10 2000000 normal 1 100000 sw-sat4j-10000-to-11000-randomized-train-first100.txt params.txt 1 1 0 2"
#	out.puts "ruby ../scripts/explore_new_focused_ils.rb sat4j median_runlength adj_mean 1 10 2000000 normal 1 100000 sw-sat4j-10000-to-11000-randomized-train-first100.txt params.txt 1 1 0 2"
#	puts "ruby focused_ils.rb branin_function median 100 10 _ test 10 100 _ 0 1" 
end

	
# =========================================
#                  MAIN
# =========================================

@eps = 0.000001
$statsfile = "jobserver_paramils_stats_#{rand}.txt"

#===  Create outfile
t = Time.now
now= t.strftime("%Y-%m-%dat%I-%M%p")

# =========================================
# ParamILS defaults.
# =========================================
@approach = "focused"
@init_def = 1
@algo = "saps"
@outdir = ""
@numRun = ""
@instance_seed_file = ""
@test_instance_seed_file = ""

@run_obj = "runtime"
@overall_obj = "adj_mean"

@N = 100
@cutoff_time = 10
@cutoff_length = "max"
@maxEvals = 20000
@tunerTimeout = 18000
@maxIts = 20000

@db = 1
@jobserver = 0

@param_file = "params.txt"
@fix_input = ""
start_input = ""

@output_level = 50
@test = false

@R = 10
@pertubation_strength_basic = 3
@pertubation_strength_scaling = 0
@p_restart = 0.01
@maxbonus = 100000

# =========================================
# Read in ParamILS-specific command line options.
# =========================================
0.step(ARGV.length-1, 2){|i|
	case ARGV[i]
	when "-instance_seed_file"
		@instance_seed_file = ARGV[i+1]
	when "-test_instance_seed_file"
		@test_instance_seed_file = ARGV[i+1]
	when "-test"
		@test = ARGV[i+1]
		@test = false if @test == "0" or @test == "false" or @test == 0
	when "-approach"
		@approach = ARGV[i+1]
	when "-numRun"
		@numRun = ARGV[i+1].to_i
	when "-outdir"
		@outdir = ARGV[i+1]

	when "-algo"
		@algo = ARGV[i+1]
	when "-run_obj"
		@run_obj = ARGV[i+1]
	when "-overall_obj"
		@overall_obj = ARGV[i+1]

	when "-cutoff_time"
		@cutoff_time = ARGV[i+1]
	when "-cutoff_length"
		@cutoff_length = ARGV[i+1]
	when "-maxEvals"
		tmp = ARGV[i+1]
		tmp = 100000000000000000 if tmp == "max"
		@maxEvals = tmp.to_i
	when "-maxIts"
		tmp = ARGV[i+1]
		tmp = 100000000000000000 if tmp == "max"
		@maxIts = tmp.to_i
	when "-tunerTimeout"
		tmp = ARGV[i+1]
		tmp = 100000000000000000 if tmp == "max"
		@tunerTimeout = tmp.to_i

#=== tuning parameter
	when "-maxbonus"
		@maxbonus = ARGV[i+1].to_i
	when "-pert_strength_basic"
		@pertubation_strength = ARGV[i+1].to_i
	when "-pert_strength_scaling"
		@pertubation_strength_scaling = ARGV[i+1].to_i
	when "-rw_prob"
		@p_restart = ARGV[i+1].to_f
	when "-R"
		@R = ARGV[i+1].to_i
	when "-init"
		if ARGV[i+1].split.length > 1
			#=== Start input values are at least partly user-defined, rest: default.
			start_input = ARGV[i+1]
			@init_def = 1
		else
			@init_def  = ARGV[i+1]
			@init_def = false if @init_def == "0" or @init_def == 0
		end
	when "-N"
		@N = ARGV[i+1].to_i
		
		
	when "-db"
		@db = ARGV[i+1]
	when "-jobserver"
		@jobserver = ARGV[i+1].to_i

	when "-paramFile"
		@param_file = ARGV[i+1]
	when "-fix"
		@fix_input  = ARGV[i+1]

	when "-output_level"
		@output_level = ARGV[i+1].to_i

	when "-scenario"
		@scenario = ARGV[i+1]

		case @scenario
		when "SAT4J-SW"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "sw-sat4j-5-to-10s-scrambled-first1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "sw-sat4j-5-to-10s-scrambled-last1000-det_algo/instance_list_0.txt" if @test

		when "SAT4J-SWorig"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = 5
			@cutoff_length = "max"

			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "sw-sat4j-1-to-2s-scrambled-first1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "sw-sat4j-1-to-2s-scrambled-last1000-det_algo/instance_list_0.txt" if @test

		when "SPEAR-swv"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "SWV-scrambled-first302-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat/instance_list_0.txt"

		when "SPEAR-bmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-first64-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat/instance_list_0.txt"

		when "SPEAR-cbmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-newsplit-train-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-newsplit-test-strat/instance_list_0.txt"

		when "SPEAR-obmc"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "mean1000" #"adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-first64-minisattime-ordered/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat/instance_list_0.txt"

		when "SPEAR-ibm"
			@algo = "spear1.2.1.1"

			@run_obj = "runtime"
			@overall_obj = "mean1000"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 8640000
			@maxEvals = 200000

			@instance_seed_file = "IBM_40randomsets-scrambled-first382-feasible287-strat/instance_list_#{@numRun}.txt"
			#			@instance_seed_file = "IBM_40randomsets-scrambled-first382-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "IBM_40randomsets-scrambled-last383-strat/instance_list_0.txt"

		when "SAT4J-bmc"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean1000"
			
			@cutoff_time = 10 #300 
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "BMC-scrambled-first64-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "BMC-scrambled-last64-strat-det_algo/instance_list_0.txt"

		when "SAT4J-swv"
			@algo = "sat4j"

			@run_obj = "runtime"
			@overall_obj = "mean1000"
			
			@cutoff_time = 10 #300
			@cutoff_length = "max"

			@tunerTimeout = 864000
			@maxEvals = 20000

			@instance_seed_file = "SWV-scrambled-first302-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SWV-scrambled-last302-strat-det_algo/instance_list_0.txt"

		when "SPEAR-pjava"
			@algo = "spear1.2.1"

			@run_obj = "runtime"
#			@overall_obj = "geomean"
			@overall_obj = "adj_mean"
			
			@cutoff_time = 300
			@cutoff_length = "max"

			@tunerTimeout = 100000
			@maxEvals = 20000

			@instance_seed_file = "PJAVA-scrambled-first38-strat/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "PJAVA-scrambled-last38-strat/instance_list_0.txt" if @test

		when "CPLEX-regions"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "adj_mean"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "cats-regions-goods100-bids500-det_algo/instance_list_#{@numRun}.txt"
			#@instance_seed_file = "cats-regions-goods200-bids1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "XXX/instance_list_0.txt" if @test

		when "CPLEX-200regions"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "adj_mean"

			@cutoff_time = 10
			@cutoff_length = "max"

			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "cats-regions-goods200-bids1000-det_algo/instance_list_#{@numRun}.txt"
			#@instance_seed_file = "cats-regions-goods200-bids1000-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "XXX/instance_list_0.txt" if @test
			
		when "CPLEX-miplib"
			@algo = "cplex"

			@run_obj = "runtime"
			@overall_obj = "mean1000"

			@cutoff_time = 100
			@cutoff_length = "max"

			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "miplib3-scrambled-first33-strat-det_algo/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "miplib3-scrambled-last32-strat-det_algo/instance_list_0.txt" if @test
		
		when "SAPS-SW"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = "10"
			@cutoff_length = "max"
			
			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "SW-saps-5s-to-10s-scrambled-first113/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-5s-to-10s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-SWorig"
			@algo = "saps"

			@run_obj = "runlength"
			@overall_obj = "median"

			@cutoff_time = "max"
			@cutoff_length = "2000000"
			
			@tunerTimeout = 36000
			@maxEvals = 20000

			@instance_seed_file = "SW-saps-1M-to2M-scrambled-first100/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-1M-to2M-scrambled-last82/instance_list_0.txt" if @test

		when "SAPS-SW100"
			@algo = "saps"

			@run_obj = "runtime"
			@overall_obj = "median"

			@cutoff_time = "max"
			@cutoff_length = 2000000
			
			@tunerTimeout = 18000
			@maxEvals = 20000

			@instance_seed_file = "SW-saps-100s-to200s-scrambled-first231/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "SW-saps-100s-to200s-scrambled-last100/instance_list_0.txt" if @test

		when "SAPS-QWH"
			@algo = "saps"

			@run_obj = "runlength"
#			@overall_obj = "q90"
			@overall_obj = "median"

			@cutoff_time = 1
			@cutoff_length = "max"
			
			@tunerTimeout = 9000
			@maxEvals = 20000

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "SAPS-QWHorig"
			@algo = "saps"

			@run_obj = "runlength"
			@overall_obj = "q90"

			@cutoff_time = 1
			@cutoff_length = "max"
			
			@tunerTimeout = 9000
			@maxEvals = 20000

			@instance_seed_file = "qwh_med_inst/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "qwh_med_inst_test/instance_list_0.txt" if @test

		when "GLS4MPE"
			@algo = "gls4mpe"

			@run_obj = "prob"
			@overall_obj = "mean"

			@cutoff_time = 10
			@cutoff_length = "max"
			
			@tunerTimeout = 200000
			@maxEvals = 20000

			@instance_seed_file = "gridGH25GW25K2-pseudo-opt-scrambled-first50/instance_list_#{@numRun}.txt"
			@test_instance_seed_file = "gridGH25GW25K2-pseudo-opt-scrambled-last50/instance_list_0.txt" if @test
		else
			raise "Undefined scenario #{@scenario}"
		end

	else
		output_help($stdout)
		puts "\n\n"
		raise "Unknown argument: #{ARGV[i]}"
	end
}

if ARGV.length < 1
	output_help($stdout)
	exit -1
end

@db = false if @db == "0" or @db  == "false" or @db  == 0

@cutoff_time = 1000000000 if @cutoff_time == "max"
@cutoff_time = @cutoff_time.to_f

@cutoff_length = 2147483647 if @cutoff_length == "max"
@cutoff_length = @cutoff_length.to_i

@oncluster = @jobserver

if @instance_seed_file == ""
	output_help($stdout)
	puts "\n\n"
	raise "Need to specify -instance_list <instance_list>" 
end

if @numRun == ""
	output_help($stdout)
	puts "\n\n"
	raise "Need to specify #{-numRun} <numRun>" 
end

if @outdir == ""
	@outdir, tmp = File.split(@instance_seed_file)
end

# =========================================
# Initial setup.
# =========================================
require "dbi_ils_accessor.rb"  if @db
$total_cputime = 0

#=== Read params from param file.
@params, @domain, @default, $conditionals, @forbidden_combos = read_params(@param_file) # Need conditionals as global to enable strip_state in param_reader
num_params = 0
for param in @params
	num_params += 1 if @domain[param].length>1
end

@pertubation_strength = @pertubation_strength_basic
if @pertubation_strength_scaling > 0
	@pertubation_strength = [@pertubation_strength_basic, (@pertubation_strength_basic*num_params/10.0).floor].max
end

@out = File.open("#{@outdir}/#{@approach}-maxbonus#{@maxbonus}-log-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-maxEval#{@maxEvals}-maxIts#{@maxIts}-init#{@init_def}-time#{@cutoff_time}-length#{@cutoff_length}_#{@numRun}.txt", "w")
@param_ils_traj_file = File.open("#{@outdir}/#{@approach}-maxbonus#{@maxbonus}-traj-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-maxEval#{@maxEvals}-maxIts#{@maxIts}-init#{@init_def}-time#{@cutoff_time}-length#{@cutoff_length}_#{@numRun}.txt", "w")
test_output_filename = "#{@outdir}/#{@approach}-maxbonus#{@maxbonus}-algo#{@algo}-runobj#{@run_obj}-overallobj#{@overall_obj}-runs#{@N}-maxEval#{@maxEvals}-maxIts#{@maxIts}-init#{@init_def}-time#{@cutoff_time}-length#{@cutoff_length}_#{@numRun}_test.txt"

string_arguments = ARGV.map{|x| "\"#{x}\""}.join(" ")
#string_arguments = ARGV.join(" ")
@out.puts "Call: /usr/bin/ruby ../scripts/param_ils_1_0_run.rb #{string_arguments}\n\n\n"
#output_help(@out)
@out.flush

#=== Build fixed parameters.
@fixed_ass = set_fixed_params(@fix_input, @domain, @params, @out)
@start_ass = set_fixed_params(start_input, @domain, @params, @out)

@number_of_params_to_opt = @params.length - @fixed_ass.keys.length

#=== Build @instances from file
@instances = []
File.open(@instance_seed_file){|file| 
    while line=file.gets;
	raise "Incorrect number of entries #{line.strip.split.length} in #{line}" unless line.strip.split.length >= 3  
	entries = line.strip.split
	seed = entries[0].to_i
	inst = entries[1]
	desired_qual = entries[2]
	reference = entries[3]
	@instances << {"seed"=>seed, "name"=>inst, "qual"=>desired_qual, "reference"=>reference, "resultForState"=>Hash.new}
    end
}

#=== Split into training and test.
validN = 1000
@test_instances = []
if @test_instance_seed_file == ""
	#=== N for both train and "test"
	@test_instances = @instances[@N...@instances.length]
	validN = 100
else 
	#=== N new ones for test
	@test_instances = []
	File.open(@test_instance_seed_file){|file| 
	    while line=file.gets;
		raise "Incorrect number of entries #{line.strip.split.length} in #{line}" unless line.strip.split.length >= 3  
		entries = line.strip.split
		seed = entries[0].to_i
		inst = entries[1]
		desired_qual = entries[2]
		reference = entries[3]
		@test_instances << {"seed"=>seed, "name"=>inst, "qual"=>desired_qual, "reference"=>reference, "resultForState"=>Hash.new}
	    end
	    validN = 100
	}
end
@instances = @instances[0...@N]
@test_instances = @test_instances[0...validN]


#=== Initialize variables.
seed = (@numRun+1)*1234 # Seed for Param_ILS, not the one to pass on to algorithms.
puts "seed: #{seed}"
@out.puts "seed: #{seed}"

@out.puts "algo: #{@algo}"
@out.puts "tunerTimeout: #{@tunerTimeout}"
@out.puts "maxEvals: #{@maxEvals}"

@out.puts "run_obj: #{@run_obj}"
@out.puts "overall_obj: #{@overall_obj}"

@out.puts "instance_seed_file: #{@instance_seed_file}"
@out.puts "test_instance_seed_file: #{@test_instance_seed_file}" unless @test_instance_seed_file == ""
@out.puts "N: #{@N}"

@out.puts "cutoff_time: #{@cutoff_time}"
@out.puts "cutoff_length: #{@cutoff_length}"

@out.puts "R: #{@R}"
@out.puts "pertubation_strength_basic: #{@pertubation_strength_basic}"
@out.puts "pertubation_strength_scaling: #{@pertubation_strength_scaling}"
@out.puts "p_restart: #{@p_restart}"
@out.puts "maxbonus: #{@maxbonus}"

srand(seed)

@iteration = 1
@numFlip = 1

@lastOutputCPUTime = 0
@evaluation_count = 0
@cachedObjectives = Hash.new
@cachedRunObjectives = Hash.new
	
@incumbent_state = nil
$total_cputime = 0
	
puts("Run #{@numRun+1}")
@out.puts("Run #{@numRun+1}")
@param_ils_traj_file.puts "Run #{@numRun+1}"

#=== Start the search for this run of ParamILS.
output "========================================================\nStarting ILS \n========================================================"
if @approach == "random"
	random_search()
else
	iterated_local_search()
end
output_incumbent_state(@incumbent_state) 

@out.close
@param_ils_traj_file.close

#=== Evaluate test/validation performance.
puts "\nComputing validation result ..."
test_results = getRunsOnEntriesWithParams(@test_instances, @incumbent_state, true)
test_result = combinationOfObjectiveFunctions(@algo, @overall_obj, test_results, @run_obj, @cutoff_time, @cutoff_length)
File.open(test_output_filename, "w"){|file|
	file.puts test_result
}
puts "Validation result: #{test_result}"